/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/audioio.h>

#include "m6809.h"
#include "pia.h"
#include "sound.h"
#include "logging.h"
#include "types.h"

#define SAMPLE_RATE 22050
#define CHANNELS 1
#define BUFFER_SIZE 512
#define SAMPLE_CYCLES (OSCILLATOR_RATE / SAMPLE_RATE)
//#define BUFFER_CYCLES ((OSCILLATOR_RATE * BUFFER_SIZE) / SAMPLE_RATE)
#define BUFFER_CYCLES ((OSCILLATOR_RATE / SAMPLE_RATE) * BUFFER_SIZE)

static int sound_fd;
static int channels = CHANNELS;
static uint_t samples_start;

static Uint32 sound_cycle_base;
static Uint8 *buffer;
static Uint8 *buffer_end;
static Uint8 *wrptr;
static Uint8 lastsample;

int sound_init(void) {
	int rate = SAMPLE_RATE;
	audio_info_t device_info;
	char *device = "/dev/audio";

	channels = CHANNELS;
	sound_fd = open(device, O_WRONLY);
	if (sound_fd == -1) {
		LOG_ERROR("Couldn't open audio device %s: %s!\n", device, strerror(errno));
		exit(0);
	}
	AUDIO_INITINFO(&device_info);
	device_info.play.sample_rate = SAMPLE_RATE;
	device_info.play.channels = CHANNELS;
	device_info.play.precision = 8;
	device_info.play.encoding = AUDIO_ENCODING_LINEAR;
	if (ioctl(sound_fd, AUDIO_SETINFO, &device_info) < 0) {
		LOG_ERROR("Failed to configure audio device %s: %s",
				device, strerror(errno));
		return 0;
	}
	if (device_info.play.channels != channels) {
		LOG_ERROR("Couldn't set desired (%d) audio channels.  Got %d.\n", channels, device_info.play.channels);
		return 0;
	}
	if (device_info.play.encoding != AUDIO_ENCODING_LINEAR) {
		LOG_ERROR("Couldn't set desired audio format.\n");
		return 0;
	}
	if (device_info.play.sample_rate != rate) {
		LOG_ERROR("Couldn't set desired (%dHz) audio rate.  Got %dHz.\n", rate, device_info.play.sample_rate);
		return 0;
	}
	LOG_DEBUG(0,"Set up audio device at %dHz, %d channels.\n", rate, channels);
	samples_start = device_info.play.samples;
	PIA_1A.wrfunc[0] = sound_write_byte;
	buffer = (Uint8 *)malloc(BUFFER_SIZE);
	memset(buffer, 0x80, BUFFER_SIZE);
	buffer_end = buffer + BUFFER_SIZE;
	wrptr = buffer;
	lastsample = 0x00;
	sound_cycle_base = 0;
	write(sound_fd, buffer, BUFFER_SIZE);
	return 1;
}

void sound_reset(void) {
}

void sound_write_byte(Uint32 octet) {
	Uint32 elapsed_cycles = current_cycle - sound_cycle_base;
	Uint32 fill_with;
	Uint8 *fill_to;
	audio_info_t device_info;
	if (elapsed_cycles > BUFFER_CYCLES * 2) {
		LOG_WARN("sound_cycle_base = %x\tcurrent_cycle = %x\t%x\n", sound_cycle_base, current_cycle, BUFFER_CYCLES);
		sound_cycle_base = current_cycle;
		elapsed_cycles = 0;
	}
	fill_to = buffer + (elapsed_cycles/(Uint32)SAMPLE_CYCLES);
	fill_with = lastsample;
	octet >>= 1;
	if (!(PIA_1B.control_register & 0x08) || PIA_0B.control_register & 0x08 || PIA_0A.control_register & 0x08)
		octet = lastsample; // 0x80;
	while (wrptr < fill_to) {
		while (wrptr < fill_to && wrptr < buffer_end)
			*(wrptr++) = fill_with;
		if (wrptr >= buffer_end) {
			do {
				ioctl(sound_fd, AUDIO_GETINFO, &device_info);
			} while ((int)(device_info.play.samples - samples_start) < (BUFFER_SIZE >> 1));
			samples_start += BUFFER_SIZE;
			write(sound_fd, buffer, BUFFER_SIZE);
			fill_to -= BUFFER_SIZE;
			sound_cycle_base += BUFFER_CYCLES;
			wrptr = buffer;
		}
	}
	lastsample = octet;
	return;
}
